def decodeRPCErrorMsg(e):
    """
    Helper function to decode the raised Exception and give it a python Exception class.

    Error codes can be obtained from libraries/protocol/include/golos/protocol/exceptions.hpp
    """
    lines = str(e).strip("\n").split("\n")
    return lines[-1]


class RPCError(Exception):
    pass


class RPCErrorRecoverable(RPCError):
    pass


class NumRetriesReached(Exception):
    pass


class InvalidNodeSchemes(Exception):
    pass


class NoAccessApi(RPCError):
    pass


class AlreadyTransactedThisBlock(RPCError):
    pass


class VoteWeightTooSmall(RPCError):
    pass


class OnlyVoteOnceEvery3Seconds(RPCError):
    pass


class AlreadyVotedSimilarily(RPCError):
    pass


class NoMethodWithName(RPCError):
    pass


class PostOnlyEvery5Min(RPCError):
    pass


class DuplicateTransaction(RPCError):
    pass


class MissingRequiredPostingAuthority(RPCError):
    pass


class UnhandledRPCError(RPCError):
    pass


class ExceededAllowedBandwidth(RPCError):
    pass


class AccountExistsException(Exception):
    pass


class AccountDoesNotExistsException(Exception):
    pass


class InsufficientAuthorityError(Exception):
    pass


class MissingKeyError(Exception):
    pass


class BlockDoesNotExistsException(Exception):
    pass


class WitnessDoesNotExistsException(Exception):
    pass


class InvalidKeyFormat(Exception):
    pass


class NoWallet(Exception):
    pass


class InvalidWifError(Exception):
    pass


class WalletExists(Exception):
    pass


class PostDoesNotExist(Exception):
    pass


class VotingInvalidOnArchivedPost(Exception):
    pass


class ReadLockFail(Exception):
    pass



DATABASE_API = "database_api"
SOCIAL_NETWORK_API = "social_network"
FOLLOW_API = "follow"
MARKET_HISTORY_API = "market_history"
PRIVATE_MESSAGE_API = "private_message"
ACCOUNT_BY_KEY_API = "account_by_key"
NETWORK_BROADCAST_API = "network_broadcast_api"
RAW_BLOCK_API = "raw_block"
BLOCK_INFO_API = "block_info"
WITNESS_API = "witness_api"
ACCOUNT_HISTORY_API = "account_history"
OPERATION_HISTORY_API = "operation_history"
TAGS_API = "tags"

API_LIST = (
    DATABASE_API,
    SOCIAL_NETWORK_API,
    FOLLOW_API,
    MARKET_HISTORY_API,
    PRIVATE_MESSAGE_API,
    ACCOUNT_BY_KEY_API,
    NETWORK_BROADCAST_API,
    RAW_BLOCK_API,
    BLOCK_INFO_API,
    WITNESS_API,
    ACCOUNT_HISTORY_API,
    OPERATION_HISTORY_API,
)



import random

from golos.amount import Amount
from golos.instance import shared_steemd_instance
from golosbase import operations, transactions
from golosbase.storage import configStorage as config


class Dex(object):
    """
    This class allows to access calls specific for the internal exchange of STEEM.

    :param Steemd steemd_instance: Steemd() instance to use when accessing a RPC
    """

    assets = ["GOLOS", "GBG"]

    def __init__(self, steemd_instance=None):
        self.steemd = steemd_instance or shared_steemd_instance()

    def _get_asset(self, symbol):
        """
        Return the properties of the assets tradeable on the network.

        :param str symbol: Symbol to get the data for (i.e. STEEM, SBD, VESTS)
        """
        if symbol == "GOLOS":
            return {"symbol": "GOLOS", "precision": 3}
        elif symbol == "GBG":
            return {"symbol": "GBG", "precision": 3}
        elif symbol == "GESTS":
            return {"symbol": "GESTS", "precision": 6}
        else:
            return None

    def _get_assets(self, quote):
        """
        Given the `quote` asset, return base.

        If quote is GBG, then base is STEEM and vice versa.
        """
        assets = self.assets.copy()
        assets.remove(quote)
        base = assets[0]
        return self._get_asset(quote), self._get_asset(base)

    def get_ticker(self):
        """
        Returns the ticker for all markets.

        Output Parameters:

        * ``latest``: Price of the order last filled
        * ``lowest_ask``: Price of the lowest ask
        * ``highest_bid``: Price of the highest bid
        * ``sbd_volume``: Volume of SBD
        * ``steem_volume``: Volume of STEEM
        * ``percent_change``: 24h change percentage (in %)

        .. note::

            Market is STEEM:SBD and prices are SBD per STEEM!

        Sample Output:

        .. code-block:: js

             {'highest_bid': 0.30100226633322913,
              'latest': 0.0,
              'lowest_ask': 0.3249636958897082,
              'percent_change': 0.0,
              'sbd_volume': 108329611.0,
              'steem_volume': 355094043.0}
        """
        t = self.steemd.get_ticker()
        return {
            "highest_bid": float(t["highest_bid"]),
            "latest": float(t["latest"]),
            "lowest_ask": float(t["lowest_ask"]),
            "percent_change": float(t["percent_change"]),
            "sbd_volume": Amount(t["sbd_volume"]),
            "steem_volume": Amount(t["steem_volume"]),
        }

    def trade_history(self, time=1 * 60 * 60, limit=100):
        """
        Returns the trade history for the internal market.

        :param int time: Show the last x seconds of trades (default 1h)
        :param int limit: amount of trades to show (<100) (default: 100)
        """
        assert limit <= 100, "'limit' has to be smaller than 100"
        return self.steemd.get_trade_history(
            transactions.fmt_time_from_now(-time), transactions.fmt_time_from_now(), limit,
        )

    def market_history_buckets(self):
        return self.steemd.get_market_history_buckets()

    def market_history(
        self, bucket_seconds=60 * 5, start_age=1 * 60 * 60, end_age=0,
    ):
        """
        Return the market history (filled orders).

        :param int bucket_seconds: Bucket size in seconds (see `returnMarketHistoryBuckets()`)
        :param int start_age: Age (in seconds) of the start of the window (default: 1h/3600)
        :param int end_age: Age (in seconds) of the end of the window (default: now/0)

        Example:

        .. code-block:: js

             {'close_sbd': 2493387,
              'close_steem': 7743431,
              'high_sbd': 1943872,
              'high_steem': 5999610,
              'id': '7.1.5252',
              'low_sbd': 534928,
              'low_steem': 1661266,
              'open': '2016-07-08T11:25:00',
              'open_sbd': 534928,
              'open_steem': 1661266,
              'sbd_volume': 9714435,
              'seconds': 300,
              'steem_volume': 30088443},
        """
        return self.steemd.get_market_history(
            bucket_seconds,
            transactions.fmt_time_from_now(-start_age - end_age),
            transactions.fmt_time_from_now(-end_age),
        )

    def buy(self, amount, quote_symbol, rate, expiration=7 * 24 * 60 * 60, killfill=False, account=None, order_id=None):
        """
        Places a buy order in a given market (buy ``quote``, sell ``base`` in market ``quote_base``). If successful, the
        method will return the order creating (signed) transaction.

        :param number amount: Amount of ``quote`` to buy
        :param str quote_symbol: GOLOS, or GBG
        :param float price: price denoted in ``base``/``quote``
        :param number expiration: (optional) expiration time of the order in seconds (defaults to 7 days)
        :param bool killfill: flag that indicates if the order shall be killed if it is not filled (defaults to False)
        :param str account: (optional) the source account for the transfer if not ``default_account``
        :param int order_id: (optional) a 32bit orderid for tracking of the created order (random by default)

        Prices/Rates are denoted in 'base', i.e. the STEEM:SBD market
        is priced in SBD per STEEM.
        """
        if not account:
            if "default_account" in config:
                account = config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")

        # We buy quote and pay with base
        quote, base = self._get_assets(quote=quote_symbol)
        op = operations.LimitOrderCreate(
            **{
                "owner": account,
                "orderid": order_id or random.getrandbits(32),
                "amount_to_sell": "{:.{prec}f} {asset}".format(
                    amount * rate, prec=base["precision"], asset=base["symbol"]
                ),
                "min_to_receive": "{:.{prec}f} {asset}".format(amount, prec=quote["precision"], asset=quote["symbol"]),
                "fill_or_kill": killfill,
                "expiration": transactions.fmt_time_from_now(expiration),
            }
        )
        return self.steemd.commit.finalizeOp(op, account, "active")

    def sell(self, amount, quote_symbol, rate, expiration=7 * 24 * 60 * 60, killfill=False, account=None, orderid=None):
        """
        Places a sell order in a given market (sell ``quote``, buy ``base`` in market ``quote_base``). If successful,
        the method will return the order creating (signed) transaction.

        :param number amount: Amount of ``quote`` to sell
        :param str quote_symbol: STEEM, or SBD
        :param float price: price denoted in ``base``/``quote``
        :param number expiration: (optional) expiration time of the order in seconds (defaults to 7 days)
        :param bool killfill: flag that indicates if the order shall be killed if it is not filled (defaults to False)
        :param str account: (optional) the source account for the transfer if not ``default_account``
        :param int orderid: (optional) a 32bit orderid for tracking of the created order (random by default)

        Prices/Rates are denoted in 'base', i.e. the STEEM:SBD market
        is priced in SBD per STEEM.
        """
        if not account:
            if "default_account" in config:
                account = config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")
        # We buy quote and pay with base
        quote, base = self._get_assets(quote=quote_symbol)
        op = operations.LimitOrderCreate(
            **{
                "owner": account,
                "orderid": orderid or random.getrandbits(32),
                "amount_to_sell": "{:.{prec}f} {asset}".format(amount, prec=quote["precision"], asset=quote["symbol"]),
                "min_to_receive": "{:.{prec}f} {asset}".format(
                    amount * rate, prec=base["precision"], asset=base["symbol"]
                ),
                "fill_or_kill": killfill,
                "expiration": transactions.fmt_time_from_now(expiration),
            }
        )
        return self.steemd.commit.finalizeOp(op, account, "active")

    def cancel(self, orderid, account=None):
        """
        Cancels an order you have placed in a given market.

        :param int orderid: the 32bit orderid
        :param str account: (optional) the source account for the transfer if not ``default_account``
        """
        if not account:
            if "default_account" in config:
                account = config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")

        op = operations.LimitOrderCancel(**{"owner": account, "orderid": orderid,})
        return self.steemd.commit.finalizeOp(op, account, "active")



__all__ = [
    "account",
    "base58",
    "bip38",
    "chains",
    "exceptions",
    "http_client",
    "operationids",
    "operations",
    "storage",
    "transactions",
    "types",
    # 'memo',
]



import datetime
import time
from contextlib import suppress

from funcy import get_in, rpartial, take, walk_values
from toolz import dissoc

from golos.amount import Amount
from golos.blockchain import Blockchain
from golos.converter import Converter
from golos.instance import shared_steemd_instance
from golos.utils import json_expand, parse_time
from golosbase.exceptions import AccountDoesNotExistsException


class Account(dict):
    """
    This class allows to easily access Account data.

    :param str account_name: Name of the account
    :param Steemd steemd_instance: Steemd() instance to use when
        accessing a RPC
    """

    def __init__(self, account_name, steemd_instance=None):
        self.steemd = steemd_instance or shared_steemd_instance()
        self.name = account_name

        # caches
        self._converter = None

        self.refresh()

    def refresh(self):
        account = self.steemd.get_account(self.name)
        if not account:
            raise AccountDoesNotExistsException

        # load json_metadata
        account = json_expand(account, "json_metadata")
        super(Account, self).__init__(account)

    def __getitem__(self, key):
        return super(Account, self).__getitem__(key)

    def items(self):
        return super(Account, self).items()

    @property
    def converter(self):
        if not self._converter:
            self._converter = Converter(self.steemd)
        return self._converter

    @property
    def profile(self):
        with suppress(TypeError):
            return get_in(self, ["json_metadata", "profile"], default={})

    @property
    def sp(self):
        vests = Amount(self["vesting_shares"]).amount
        return round(self.converter.vests_to_sp(vests), 3)

    @property
    def rep(self):
        raise DeprecationWarning("Reputations field was removed from result struct")

    @property
    def balances(self):
        return self.get_balances()

    def get_balances(self):
        available = {
            "GOLOS": Amount(self["balance"]).amount,
            "GBG": Amount(self["sbd_balance"]).amount,
            "GESTS": Amount(self["vesting_shares"]).amount,
        }

        savings = {
            "GOLOS": Amount(self["savings_balance"]).amount,
            "GBG": Amount(self["savings_sbd_balance"]).amount,
        }

        accumulative = {"GOLOS": Amount(self["accumulative_balance"]).amount}

        tip = {"GOLOS": Amount(self["tip_balance"]).amount}

        totals = {
            "GOLOS": sum([available["GOLOS"], savings["GOLOS"], accumulative["GOLOS"], tip["GOLOS"]]),
            "GBG": sum([available["GBG"], savings["GBG"]]),
            "GESTS": sum([available["GESTS"]]),
        }

        total = walk_values(rpartial(round, 3), totals)

        return {
            "available": available,
            "savings": savings,
            "accumulative": accumulative,
            "tip": tip,
            "total": total,
        }

    def voting_power(self):
        return self["voting_power"] / 100

    def get_followers(self, limit: int = None, offset: str = None):
        return [x["follower"] for x in self._get_followers(direction="follower", limit=limit, offset=offset)]

    def get_following(self, limit: int = None, offset: str = None):
        return [x["following"] for x in self._get_followers(direction="following", limit=limit, offset=offset)]

    def _get_followers(self, direction="follower", limit=None, offset=""):
        users = []

        get_users = {"follower": self.steemd.get_followers, "following": self.steemd.get_following}[direction]

        limit = limit or 10 ** 6
        max_request_limit = 100
        left_number = limit

        while left_number > 0:
            select_limit = min(left_number, max_request_limit)
            result = get_users(self.name, offset, "blog", select_limit)
            users.extend(result)

            has_next = len(users) < limit and len(result) >= select_limit
            if has_next:
                if users:
                    del users[-1]
                offset = result[-1][direction]

                left_number = left_number - len(result) + 1
            else:
                left_number = 0

        return users

    def has_voted(self, post):
        active_votes = {v["voter"]: v for v in getattr(post, "active_votes")}
        return self.name in active_votes

    def curation_stats(self):
        trailing_24hr_t = time.time() - datetime.timedelta(hours=24).total_seconds()
        trailing_7d_t = time.time() - datetime.timedelta(days=7).total_seconds()

        reward_24h = 0.0
        reward_7d = 0.0

        for reward in take(5000, self.history_reverse(filter_by="curation_reward")):

            timestamp = parse_time(reward["timestamp"]).timestamp()
            if timestamp > trailing_7d_t:
                reward_7d += Amount(reward["reward"]).amount

            if timestamp > trailing_24hr_t:
                reward_24h += Amount(reward["reward"]).amount

        reward_7d = self.converter.vests_to_sp(reward_7d)
        reward_24h = self.converter.vests_to_sp(reward_24h)
        return {
            "24hr": reward_24h,
            "7d": reward_7d,
            "avg": reward_7d / 7,
        }

    def virtual_op_count(self):
        try:
            last_item = self.steemd.get_account_history(self.name, -1, 0)[0][0]
        except IndexError:
            return 0
        else:
            return last_item

    def get_account_votes(self):
        return self.steemd.get_account_votes(self.name)

    def get_withdraw_routes(self):
        return self.steemd.get_withdraw_routes(self.name, "all")

    def get_conversion_requests(self):
        return self.steemd.get_conversion_requests(self.name)

    @staticmethod
    def filter_by_date(items, start_time, end_time=None):
        start_time = parse_time(start_time).timestamp()
        if end_time:
            end_time = parse_time(end_time).timestamp()
        else:
            end_time = time.time()

        filtered_items = []
        for item in items:
            item_time = None
            if "time" in item:
                item_time = item["time"]
            elif "timestamp" in item:
                item_time = item["timestamp"]

            if item_time:
                timestamp = parse_time(item_time).timestamp()
                if end_time > timestamp > start_time:
                    filtered_items.append(item)

        return filtered_items

    def export(self, load_extras=True):
        """
        This method returns a dictionary that is type-safe to store as JSON or in a database.

        :param bool load_extras: Fetch extra information related to the account (this might take a while).
        """
        extras = dict()
        if load_extras:
            followers = self.get_followers()
            following = self.get_following()
            extras = {
                "followers": followers,
                "followers_count": len(followers),
                "following": following,
                "following_count": len(following),
                "curation_stats": self.curation_stats(),
                "withdrawal_routes": self.get_withdraw_routes(),
                "conversion_requests": self.get_conversion_requests(),
            }

        composed_dict = self.copy()
        composed_dict.update(extras)
        composed_dict.update(
            {"profile": self.profile, "sp": self.sp, "balances": self.get_balances(),}
        )

        return composed_dict

    def get_account_history(self, index, limit, start=None, stop=None, order=-1, filter_by=None, raw_output=False):
        """
        A generator over steemd.get_account_history.

        It offers serialization, filtering and fine grained iteration control.

        Args:
            index (int): start index for get_account_history
            limit (int): How many items in account history will be scanned (any ops, not only filtered)
            start (int): (Optional) skip items until this index
            stop (int): (Optional) stop iteration early at this index
            order: (1, -1): 1 for chronological, -1 for reverse order
            filter_by (str, list): filter out all but these operations
            raw_output (bool): (Defaults to False). If True, return history in
                steemd format (unchanged).
        """
        history = self.steemd.get_account_history(self.name, index, limit)
        for item in history[::order]:
            index, event = item

            # start and stop utilities for chronological generator
            if start and index < start:
                continue

            if stop and index > stop:
                return

            op_type, op = event["op"]
            block_props = dissoc(event, "op")

            def construct_op(account_name):
                # verbatim output from steemd
                if raw_output:
                    return item

                # index can change during reindexing in
                # future hard-forks. Thus we cannot take it for granted.
                immutable = op.copy()
                immutable.update(block_props)
                immutable.update(
                    {"account": account_name, "type": op_type,}
                )
                _id = Blockchain.hash_op(immutable)
                immutable.update(
                    {"_id": _id, "index": index,}
                )
                return immutable

            if filter_by is None:
                yield construct_op(self.name)
            else:
                if type(filter_by) is list:
                    if op_type in filter_by:
                        yield construct_op(self.name)

                if type(filter_by) is str:
                    if op_type == filter_by:
                        yield construct_op(self.name)

    def history(self, filter_by=None, start=0, batch_size=1000, raw_output=False):
        """Stream account history in chronological order."""
        max_index = self.virtual_op_count()
        if not max_index:
            return

        start_index = start + batch_size
        i = start_index
        while i < max_index + batch_size:
            yield from self.get_account_history(
                index=i,
                limit=batch_size,
                start=i - batch_size,
                stop=max_index,
                order=1,
                filter_by=filter_by,
                raw_output=raw_output,
            )
            i += batch_size + 1

    def history_reverse(self, filter_by=None, batch_size=1000, raw_output=False):
        """Stream account history in reverse chronological order."""
        start_index = self.virtual_op_count()
        if not start_index:
            return

        i = start_index
        while i > 0:
            if i - batch_size < 0:
                batch_size = i
            yield from self.get_account_history(
                index=i, limit=batch_size, order=-1, filter_by=filter_by, raw_output=raw_output,
            )
            i -= batch_size + 1

    def rawhistory(self, first=99999999999, limit=-1, only_ops=[], exclude_ops=[]):
        """
        Returns a generator for individual account transactions. The latest operation will be first. This call can be
        used in a ``for`` loop.

        :param str account: account name to get history for
        :param int first: sequence number of the first transaction to return
        :param int limit: limit number of filtered operations to return
        :param array only_ops: Limit generator by these operations
        """
        cnt = 0
        _limit = 100
        if _limit > first:
            _limit = first
        while first > 0:
            # RPC call
            txs = self.steemd.get_account_history(self.name, first, _limit)
            for i in txs[::-1]:
                if exclude_ops and i[1]["op"][0] in exclude_ops:
                    continue
                if not only_ops or i[1]["op"][0] in only_ops:
                    cnt += 1
                    yield i
                    if limit >= 0 and cnt >= limit:
                        break
            if limit >= 0 and cnt >= limit:
                break
            if len(txs) < _limit:
                break
            first = txs[0][0] - 1  # new first
            if _limit > first:
                _limit = first



from golos.commit import Commit
from golos.consts import API_LIST
from golos.steemd import Steemd


class Steem:
    """
    Connect to the Steem network.

    Args:

        nodes (list): A list of Steem HTTP RPC nodes to connect to. If
        not provided, official Steemit nodes will be used.

        debug (bool): Elevate logging level to `logging.DEBUG`.
        Defaults to `logging.INFO`.

        no_broadcast (bool): If set to ``True``, committal actions like
        sending funds will have no effect (simulation only).

        num_retries (int): Limit connection attempts to the node. Defaults to -1
        (unlimited).

    Optional Arguments (kwargs):

    Args:

        keys (list): A list of wif keys. If provided, the Wallet will
        use these keys rather than the ones found in BIP38 encrypted
        wallet.

        unsigned (bool): (Defaults to False) Use this for offline signing.

        expiration (int): (Defualts to 60) Size of window in seconds
        that the transaction needs to be broadcasted in, before it
        expires.

    Returns:

        Steemd class instance. It can be used to execute commands
        against golos node.

    Example:

       If you would like to override the official Steemit nodes
       (default), you can pass your own.  When currently used node goes
       offline, ``Steemd`` will automatically fail-over to the next
       available node.

       .. code-block:: python

           nodes = [
               'https://steemd.yournode1.com',
               'https://steemd.yournode2.com',
           ]

           s = Steemd(nodes)
    """

    def __init__(self, nodes=None, no_broadcast=False, **kwargs):
        self.steemd = Steemd(nodes=nodes, **kwargs)
        self.commit = Commit(steemd_instance=self.steemd, no_broadcast=no_broadcast, **kwargs)

    def __getattr__(self, item):
        """Bind .commit, .steemd methods here as a convenience."""
        if hasattr(self.steemd, item):
            return getattr(self.steemd, item)
        if hasattr(self.commit, item):
            return getattr(self.commit, item)
        if item in API_LIST:
            return Steem.Api(api_name=item, exec_method=self.steemd.call)

        raise AttributeError('Steem has no attribute "%s"' % item)

    class Api(object):
        def __init__(self, api_name="", exec_method=None):
            self.api_name = api_name
            self.exec_method = exec_method
            return

        def __getattr__(self, method_name):
            return Steem.Method(api_name=self.api_name, method_name=method_name, exec_method=self.exec_method,)

    class Method(object):
        def __init__(self, api_name="", method_name="", exec_method=None):
            self.api_name = api_name
            self.method_name = method_name
            self.exec_method = exec_method
            return

        def __call__(self, *args, **kwargs):
            assert not (args and kwargs), "specified both args and kwargs"
            if len(kwargs) > 0:
                return self.exec_method(self.method_name, kwargs=kwargs, api=self.api_name)
            return self.exec_method(self.method_name, *args, api=self.api_name)


if __name__ == "__main__":
    s = Steem()
    print(s.get_account_count())



import hashlib
import json
import logging
import time
import warnings
from typing import Union

from golos.consts import DATABASE_API, OPERATION_HISTORY_API
from golos.instance import shared_steemd_instance
from golos.steemd import Steemd

logger = logging.getLogger(__name__)

virtual_operations = [
    "fill_convert_request",
    "author_reward",
    "curation_reward",
    "comment_reward",
    "liquidity_reward",
    "interest",
    "fill_vesting_withdraw",
    "fill_order",
    "shutdown_witness",
    "fill_transfer_from_savings",
    "hardfork",
    "comment_payout_update",
    "comment_benefactor_reward",
    "return_vesting_delegation",
]


class Blockchain(object):
    """
    Access the blockchain and read data from it.

    Args:
        steemd_instance (Steemd): Steemd() instance to use when accessing a RPC
        mode (str): `irreversible` or `head`. `irreversible` is default.
    """

    def __init__(self, steemd_instance=None, mode="irreversible"):
        self.steem = steemd_instance or shared_steemd_instance()
        self.config = self.steem.get_config()

        if mode == "irreversible":
            self.mode = "last_irreversible_block_num"
        elif mode == "head":
            self.mode = "head_block_number"
        else:
            raise ValueError("invalid value for 'mode'!")

    def info(self):
        """This call returns the *dynamic global properties*"""
        return self.steem.get_dynamic_global_properties()

    def get_current_block_num(self):
        """This call returns the current block."""
        return self.info().get(self.mode)

    def get_current_block(self):
        """This call returns the current block."""
        return self.steem.get_block(self.get_current_block_num())

    def stream_from(
        self,
        start_block=None,
        end_block=None,
        batch_operations=False,
        full_blocks=False,
        only_virtual_ops=False,
        **kwargs
    ):
        """
        This call yields raw blocks or operations depending on ``full_blocks`` param.

        By default, this generator will yield operations, one by one.
        You can choose to yield lists of operations, batched to contain
        all operations for each block with ``batch_operations=True``.
        You can also yield full blocks instead, with ``full_blocks=True``.

        Args:
            start_block (int): Block to start with. If not provided, current (head) block is used.
            end_block (int): Stop iterating at this block. If not provided, this generator will run forever (streaming mode).
            batch_operations (bool): (Defaults to False) Rather than yielding operations one by one,
                yield a list of all operations for each block.
            full_blocks (bool): (Defaults to False) Rather than yielding operations, return raw, unedited blocks as
                provided by steemd. This mode will NOT include virtual operations.
        """

        _ = kwargs  # we need this
        # Let's find out how often blocks are generated!
        block_interval = self.config.get("STEEMIT_BLOCK_INTERVAL") or 3

        is_reversed = end_block and start_block > end_block

        if not start_block:
            start_block = self.get_current_block_num()

        while True:
            head_block = self.get_current_block_num()

            range_params = (start_block, head_block + 1)
            if end_block is not None and start_block > end_block:
                range_params = (start_block, max(0, end_block - 2), -1)

            for block_num in range(*range_params):
                if end_block is not None:
                    if is_reversed and block_num < end_block:
                        return
                    elif not is_reversed and block_num > end_block:
                        return

                if full_blocks:
                    block = self.steem.get_block(block_num)
                    # inject block number
                    block.update({"block_num": block_num})
                    yield block
                elif batch_operations:
                    yield self.steem.get_ops_in_block(block_num, only_virtual_ops)
                else:
                    ops = self.steem.get_ops_in_block(block_num, only_virtual_ops)
                    for op in ops:
                        # avoid yielding empty ops
                        if op:
                            yield op

            # next round
            start_block = head_block + 1
            time.sleep(block_interval)

    def reliable_stream(
        self,
        start_block=None,
        block_interval=None,
        update_interval=False,
        batch_operations=False,
        full_blocks=False,
        timeout=None,
        **kwargs
    ):
        """
        A version of stream_from() intended for use in services that NEED reliable (nonstop) streaming.

        By default, works same as stream_from() but will also keep trying
        until getting a response from steemd, allowing catching up after
        server downtime.

        Warnings: To ensure reliability, this method does some weird
        none-standard things with the steemd client

        Args:

        start_block (int): Block to start with. If not provided, current
        (head) block is used.

        block_interval (int): Time between block generations. If not
        provided, will attempt to query steemd for this value

        batch_operations (bool): (Defaults to False) Rather than yielding
        operations one by one, yield a list of all operations for each
        block.

        full_blocks (bool): (Defaults to False) Rather than yielding
        operations, return raw, unedited blocks as provided by steemd. This
        mode will NOT include virtual operations.

        timeout (int): Time to wait on response from steemd before assuming
        timeout and retrying queries. If not provided, this will default to
        block_interval/4 for all queries except get_block_interval() -
        where it will default to 2 seconds for initial setup
        """

        def get_reliable_client(_timeout):
            # we want to fail fast and try the next node quickly
            return Steemd(nodes=self.steem.nodes, retries=1, timeout=_timeout, re_raise=True)

        def reliable_query(_client, _method, _api, *_args):
            # this will ALWAYS eventually return, at all costs
            while True:
                try:
                    return _client.call(_method, *_args, api=_api)
                except Exception as e:
                    logger.error(
                        "Error: %s" % str(s),
                        extra=dict(exc=e, response=retval, api_name=_api, api_method=_method, api_args=_args),
                    )
                    time.sleep(1)

        def get_reliable_block_interval(_client):
            return reliable_query(_client, "get_config", DATABASE_API).get("STEEMIT_BLOCK_INTERVAL") or 3

        def get_reliable_current_block(_client):
            return reliable_query(_client, "get_dynamic_global_properties", DATABASE_API).get(self.mode)

        def get_reliable_blockdata(_client, _block_num):
            return reliable_query(_client, "get_block", DATABASE_API, block_num)

        def get_reliable_ops_in_block(_client, _block_num):
            return reliable_query(_client, "get_ops_in_block", OPERATION_HISTORY_API, block_num, False)

        if timeout is None:
            if block_interval is None:
                _reliable_client = get_reliable_client(2)
                block_interval = get_reliable_block_interval(_reliable_client)
            else:
                timeout = block_interval / 4
                _reliable_client = get_reliable_client(timeout)
        else:
            _reliable_client = get_reliable_client(timeout)
        if block_interval is None:
            block_interval = get_reliable_block_interval(_reliable_client)
        if start_block is None:
            start_block = get_reliable_current_block(_reliable_client)

        while True:
            sleep_interval = block_interval / 4
            head_block = get_reliable_current_block(_reliable_client)

            for block_num in range(start_block, head_block + 1):
                if full_blocks:
                    yield get_reliable_current_block(_reliable_client, head_block)
                elif batch_operations:
                    yield get_reliable_ops_in_block(_reliable_client, head_block)
                else:
                    for reliable_ops in get_reliable_ops_in_block(_reliable_client, head_block):
                        yield reliable_ops

                sleep_interval = sleep_interval / 2

            time.sleep(sleep_interval)
            start_block = head_block + 1

    def stream(self, filter_by: Union[str, list] = list(), *args, **kwargs):
        """
        Yield a stream of specific operations, starting with current head block.

        This method can work in 2 modes:
        1. Whether only real operations are requested, it will use get_block() API call, so you don't need to have
        neigher operation_history nor accunt_history plugins enabled.
        2. Whether you're requesting any of the virtual operations, your node should have operation_history or
        accunt_history plugins enabled and appropriate settings for the history-related params should be set
        (history-start-block, history-whitelist-ops or history-blacklist-ops).

        The dict output is formated such that ``type`` caries the operation type, timestamp and block_num are taken
        from the block the operation was stored in and the other key depend on the actual operation.

        Args:
            start_block (int): Block to start with. If not provided, current (head) block is used.
            end_block (int): Stop iterating at this block. If not provided, this generator will run forever (streaming mode).
            filter_by (str, list): List of operations to filter for
        """
        if isinstance(filter_by, str):
            filter_by = [filter_by]

        if not bool(set(filter_by).intersection(virtual_operations)):
            # uses get_block instead of get_ops_in_block
            for block in self.stream_from(full_blocks=True, *args, **kwargs):
                for tx in block.get("transactions"):
                    for op in tx["operations"]:
                        if not filter_by or op[0] in filter_by:
                            r = {
                                "type": op[0],
                                "timestamp": block.get("timestamp"),
                                "block_num": block.get("block_num"),
                            }
                            r.update(op[1])
                            yield r
        else:
            # uses get_ops_in_block
            kwargs["only_virtual_ops"] = not bool(set(filter_by).difference(virtual_operations))
            for op in self.stream_from(full_blocks=False, *args, **kwargs):
                if kwargs.get("raw_output"):
                    yield op

                if not filter_by or op["op"][0] in filter_by:
                    r = {
                        "_id": self.hash_op(op),
                        "type": op["op"][0],
                        "timestamp": op.get("timestamp"),
                        "block_num": op.get("block"),
                        "trx_id": op.get("trx_id"),
                    }
                    r.update(op["op"][1])
                    yield r

    def history(self, filter_by: Union[str, list] = list(), start_block=1, end_block=None, raw_output=False, **kwargs):
        """
        Yield a stream of historic operations.

        Similar to ``Blockchain.stream()``, but starts at beginning of chain unless ``start_block`` is set.

        Args:
            filter_by (str, list): List of operations to filter for
            start_block (int): Block to start with. If not provided, start of blockchain is used (block 1).
            end_block (int): Stop iterating at this block. If not provided, this generator will run forever.
            raw_output (bool): (Defaults to False). If True, return ops in a unmodified steemd structure.
        """
        return self.stream(
            filter_by=filter_by, start_block=start_block, end_block=end_block, raw_output=raw_output, **kwargs
        )

    def ops(self, *args, **kwargs):
        raise DeprecationWarning("Blockchain.ops() is deprecated. Please use Blockchain.stream_from() instead.")

    def replay(self, **kwargs):
        warnings.warn("Blockchain.replay() is deprecated. Please use Blockchain.history() instead.")
        return self.history(**kwargs)

    @staticmethod
    def hash_op(event: dict):
        """This method generates a hash of blockchain operation."""
        data = json.dumps(event, sort_keys=True)
        return hashlib.sha1(bytes(data, "utf-8")).hexdigest()

    def get_all_usernames(self, *args, **kwargs):
        """Fetch the full list of GOLOS usernames."""
        _ = args, kwargs
        warnings.warn("Blockchain.get_all_usernames() is now at Steemd.get_all_usernames().")
        return self.steem.get_all_usernames()



op_names = [
    "vote",
    "comment",
    "transfer",
    "transfer_to_vesting",
    "withdraw_vesting",
    "limit_order_create",
    "limit_order_cancel",
    "feed_publish",
    "convert",
    "account_create",
    "account_update",
    "witness_update",
    "account_witness_vote",
    "account_witness_proxy",
    "pow",
    "custom",
    "report_over_production",
    "delete_comment",
    "custom_json",
    "comment_options",
    "set_withdraw_vesting_route",
    "limit_order_create2",
    "challenge_authority",
    "prove_authority",
    "request_account_recovery",
    "recover_account",
    "change_recovery_account",
    "escrow_transfer",
    "escrow_dispute",
    "escrow_release",
    "pow2",
    "escrow_approve",
    "transfer_to_savings",
    "transfer_from_savings",
    "cancel_transfer_from_savings",
    "custom_binary",
    "decline_voting_rights",
    "reset_account",
    "set_reset_account",
    "delegate_vesting_shares",
    "account_create_with_delegation",
    "account_metadata",
    "proposal_create",
    "proposal_update",
    "proposal_delete",
    "chain_properties_update",
    "break_free_referral",
    "delegate_vesting_shares_with_interest",
    "reject_vesting_shares_delegation",
    "transit_to_cyberway",
    "worker_request",
    "worker_request_delete",
    "worker_request_vote",
    "claim",
    "donate",
    "transfer_to_tip",
    "transfer_from_tip",
    "invite",
    "invite_claim",
    "account_create_with_invite",
    "fill_convert_request",
    "author_reward",
    "curation_reward",
    "comment_reward",
    "liquidity_reward",
    "interest",
    "fill_vesting_withdraw",
    "fill_order",
    "shutdown_witness",
    "fill_transfer_from_savings",
    "hardfork",
    "comment_payout_update",
    "comment_benefactor_reward",
    "return_vesting_delegation",
    "producer_reward",
    "delegation_reward",
    "auction_window_reward",
    "total_comment_reward",
    "worker_reward",
    "worker_state",
    "convert_sbd_debt",
]

#: assign operation ids
operations = dict(zip(op_names, range(len(op_names))))



import importlib
import json
import re
import struct
from collections import OrderedDict

from golosbase.account import PublicKey
from golosbase.operationids import operations
from golosbase.types import (
    Array,
    Bool,
    Bytes,
    Id,
    Int16,
    Int64,
    JsonObj,
    Map,
    OperationWrapper,
    Optional,
    PointInTime,
    StaticVariant,
    String,
    Uint16,
    Uint32,
    Uint64,
    variable_buffer,
    varint,
)

default_prefix = "GLS"

asset_precision = {
    "GOLOS": 3,
    "GESTS": 6,
    "GBG": 3,
}


class Operation:
    def __init__(self, op):
        if isinstance(op, list) and len(op) == 2:
            if isinstance(op[0], int):
                self.opId = op[0]
                name = self.get_operation_name_for_id(self.opId)
            else:
                self.opId = operations.get(op[0], None)
                name = op[0]
                if self.opId is None:
                    raise ValueError("Unknown operation")

            # convert method name like feed_publish to class name like FeedPublish
            self.name = self.to_class_name(name)
            try:
                klass = self.get_class(self.name)
            except:
                raise NotImplementedError("Unimplemented Operation %s" % self.name)
            else:
                self.op = klass(op[1])
        else:
            self.op = op
            # class name like FeedPublish
            self.name = type(self.op).__name__
            self.opId = operations[self.to_method_name(self.name)]

    @staticmethod
    def get_operation_name_for_id(_id: int):
        """Convert an operation id into the corresponding string."""
        for key, value in operations.items():
            if value == int(_id):
                return key

    @staticmethod
    def to_class_name(method_name: str):
        """Take a name of a method, like feed_publish and turn it into class name like FeedPublish."""
        return "".join(map(str.title, method_name.split("_")))

    @staticmethod
    def to_method_name(class_name: str):
        """Take a name of a class, like FeedPublish and turn it into method name like feed_publish."""
        words = re.findall("[A-Z][^A-Z]*", class_name)
        return "_".join(map(str.lower, words))

    @staticmethod
    def get_class(class_name: str):
        """Given name of a class from `operations`, return real class."""
        module = importlib.import_module("golosbase.operations")
        return getattr(module, class_name)

    def __bytes__(self):
        return bytes(Id(self.opId)) + bytes(self.op)

    def __str__(self):
        return json.dumps([self.get_operation_name_for_id(self.opId), self.op.json()])


class GrapheneObject(object):
    """
    Core abstraction class.

    This class is used for any JSON reflected object in Graphene.

    * ``instance.__json__()``: encodes data into json format
    * ``bytes(instance)``: encodes data into wire format
    * ``str(instances)``: dumps json object as string
    """

    def __init__(self, data=None):
        self.data = data

    def __bytes__(self):
        if self.data is None:
            return bytes()
        b = b""
        for name, value in self.data.items():
            if isinstance(value, str):
                b += bytes(value, "utf-8")
            else:
                b += bytes(value)
        return b

    def __json__(self):
        if self.data is None:
            return {}
        d = {}  # JSON output is *not* ordered
        for name, value in self.data.items():
            if isinstance(value, Optional) and value.isempty():
                continue

            if isinstance(value, String):
                d.update({name: str(value)})
            else:
                d.update({name: JsonObj(value)})
        return d

    def __str__(self):
        return json.dumps(self.__json__())

    def json(self):
        return self.__json__()


class Permission(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            prefix = kwargs.pop("prefix", default_prefix)

            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            # Sort keys (FIXME: ideally, the sorting is part of Public
            # Key and not located here)
            kwargs["key_auths"] = sorted(
                kwargs["key_auths"], key=lambda x: repr(PublicKey(x[0], prefix=prefix)), reverse=False,
            )
            kwargs["account_auths"] = sorted(kwargs["account_auths"], key=lambda x: x[0], reverse=False,)

            accountAuths = Map([[String(e[0]), Uint16(e[1])] for e in kwargs["account_auths"]])
            keyAuths = Map([[PublicKey(e[0], prefix=prefix), Uint16(e[1])] for e in kwargs["key_auths"]])
            super().__init__(
                OrderedDict(
                    [
                        ("weight_threshold", Uint32(int(kwargs["weight_threshold"]))),
                        ("account_auths", accountAuths),
                        ("key_auths", keyAuths),
                    ]
                )
            )


class Memo(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            prefix = kwargs.pop("prefix", default_prefix)

            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            super().__init__(
                OrderedDict(
                    [
                        ("from", PublicKey(kwargs["from"], prefix=prefix)),
                        ("to", PublicKey(kwargs["to"], prefix=prefix)),
                        ("nonce", Uint64(int(kwargs["nonce"]))),
                        ("check", Uint32(int(kwargs["check"]))),
                        ("encrypted", Bytes(kwargs["encrypted"])),
                    ]
                )
            )


class Vote(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("voter", String(kwargs["voter"])),
                        ("author", String(kwargs["author"])),
                        ("permlink", String(kwargs["permlink"])),
                        ("weight", Int16(kwargs["weight"])),
                    ]
                )
            )


class Comment(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            meta = ""
            if "json_metadata" in kwargs and kwargs["json_metadata"]:
                if isinstance(kwargs["json_metadata"], dict) or isinstance(kwargs["json_metadata"], list):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]

            super().__init__(
                OrderedDict(
                    [
                        ("parent_author", String(kwargs["parent_author"])),
                        ("parent_permlink", String(kwargs["parent_permlink"])),
                        ("author", String(kwargs["author"])),
                        ("permlink", String(kwargs["permlink"])),
                        ("title", String(kwargs["title"])),
                        ("body", String(kwargs["body"])),
                        ("json_metadata", String(meta)),
                    ]
                )
            )


class DeleteComment(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            super().__init__(
                OrderedDict([("author", String(kwargs["author"])), ("permlink", String(kwargs["permlink"])),])
            )


class Amount:
    def __init__(self, d):
        self.amount, self.asset = d.strip().split(" ")
        self.amount = float(self.amount)

        if self.asset in asset_precision:
            self.precision = asset_precision[self.asset]
        else:
            raise Exception("Asset unknown")

    def __bytes__(self):
        # padding
        asset = self.asset + "\x00" * (7 - len(self.asset))
        amount = round(float(self.amount) * 10 ** self.precision)
        return struct.pack("<q", amount) + struct.pack("<b", self.precision) + bytes(asset, "ascii")

    def __str__(self):
        return "{:.{}f} {}".format(self.amount, self.precision, self.asset)


class ExchangeRate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            super().__init__(OrderedDict([("base", Amount(kwargs["base"])), ("quote", Amount(kwargs["quote"])),]))


class ChainProperties(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            super().__init__(
                OrderedDict(
                    [
                        ("account_creation_fee", Amount(kwargs["account_creation_fee"])),
                        ("maximum_block_size", Uint32(kwargs["maximum_block_size"])),
                        ("sbd_interest_rate", Uint16(kwargs["sbd_interest_rate"])),
                    ]
                )
            )


class ChainProperties18(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            p = ChainProperties(kwargs).data
            p.update(
                OrderedDict(
                    [
                        ("create_account_min_golos_fee", Amount(kwargs["create_account_min_golos_fee"])),
                        ("create_account_min_delegation", Amount(kwargs["create_account_min_delegation"])),
                        ("create_account_delegation_time", Uint32(kwargs["create_account_delegation_time"])),
                        ("min_delegation", Amount(kwargs["min_delegation"])),
                    ]
                )
            )
            super().__init__(p)


class ChainProperties19(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            p = ChainProperties18(kwargs).data
            p.update(
                OrderedDict(
                    [
                        # Note that param order is critical to serialization
                        ("max_referral_interest_rate", Uint16(kwargs["max_referral_interest_rate"])),
                        ("max_referral_term_sec", Uint32(kwargs["max_referral_term_sec"])),
                        ("min_referral_break_fee", Amount(kwargs["min_referral_break_fee"])),
                        ("max_referral_break_fee", Amount(kwargs["max_referral_break_fee"])),
                        ("posts_window", Uint16(kwargs["posts_window"])),
                        ("posts_per_window", Uint16(kwargs["posts_per_window"])),
                        ("comments_window", Uint16(kwargs["comments_window"])),
                        ("comments_per_window", Uint16(kwargs["comments_per_window"])),
                        ("votes_window", Uint16(kwargs["votes_window"])),
                        ("votes_per_window", Uint16(kwargs["votes_per_window"])),
                        ("auction_window_size", Uint16(kwargs["auction_window_size"])),
                        ("max_delegated_vesting_interest_rate", Uint16(kwargs["max_delegated_vesting_interest_rate"])),
                        ("custom_ops_bandwidth_multiplier", Uint16(kwargs["custom_ops_bandwidth_multiplier"])),
                        ("min_curation_percent", Uint16(kwargs["min_curation_percent"])),
                        ("max_curation_percent", Uint16(kwargs["max_curation_percent"])),
                        ("curation_reward_curve", Uint64(kwargs["curation_reward_curve"])),
                        ("allow_distribute_auction_reward", Bool(bool(kwargs["allow_distribute_auction_reward"]))),
                        (
                            "allow_return_auction_reward_to_fund",
                            Bool(bool(kwargs["allow_return_auction_reward_to_fund"])),
                        ),
                    ]
                )
            )
            super().__init__(p)


class ChainProperties22(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            p = ChainProperties19(kwargs).data
            p.update(
                OrderedDict(
                    [
                        ("worker_reward_percent", Uint16(kwargs["worker_reward_percent"])),
                        ("witness_reward_percent", Uint16(kwargs["witness_reward_percent"])),
                        ("vesting_reward_percent", Uint16(kwargs["vesting_reward_percent"])),
                        ("worker_request_creation_fee", Amount(kwargs["worker_request_creation_fee"])),
                        ("worker_request_approve_min_percent", Uint16(kwargs["worker_request_approve_min_percent"])),
                        ("sbd_debt_convert_rate", Uint16(kwargs["sbd_debt_convert_rate"])),
                        ("vote_regeneration_per_day", Uint32(kwargs["vote_regeneration_per_day"])),
                        ("witness_skipping_reset_time", Uint32(kwargs["witness_skipping_reset_time"])),
                        ("witness_idleness_time", Uint32(kwargs["witness_idleness_time"])),
                        ("account_idleness_time", Uint32(kwargs["account_idleness_time"])),
                    ]
                )
            )
            super().__init__(p)


class ChainProperties23(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            p = ChainProperties22(kwargs).data
            p.update(
                OrderedDict(
                    [
                        ("claim_idleness_time", Uint32(kwargs["claim_idleness_time"])),
                        ("min_invite_balance", Amount(kwargs["min_invite_balance"])),
                    ]
                )
            )
            super().__init__(p)


class Props(StaticVariant):
    def __init__(self, o):
        type_id, data = o

        if type_id == 0:
            data = ChainProperties(data["props"])
        elif type_id == 1:
            data = ChainProperties18(data["props"])
        elif type_id == 2:
            data = ChainProperties19(data["props"])
        elif type_id == 3:
            data = ChainProperties22(data["props"])
        elif type_id == 4:
            data = ChainProperties23(data["props"])
        super().__init__(data, type_id)


class Beneficiary(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict([("account", String(kwargs["account"])), ("weight", Int16(kwargs["weight"])),])
            )


class Beneficiaries(GrapheneObject):
    def __init__(self, kwargs):
        super().__init__(OrderedDict([("beneficiaries", Array([Beneficiary(o) for o in kwargs["beneficiaries"]])),]))


class Destination(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(OrderedDict([("destination", Uint64(kwargs["destination"])),]))


class Percent(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(OrderedDict([("percent", Uint16(kwargs["percent"])),]))


class CommentOptionExtensions(StaticVariant):
    """
    Serialize Comment Payout Beneficiaries.

    Args:
        beneficiaries (list): A static_variant containing beneficiaries.

    Example:

        ::

            [0,
                {'beneficiaries': [
                    {'account': 'furion', 'weight': 10000}
                ]}
            ]
    """

    def __init__(self, o):
        type_id, data = o
        if type_id == 0:
            data = Beneficiaries(data)
        elif type_id == 1:
            data = Destination(data)
        elif type_id == 2:
            data = Percent(data)
        else:
            raise Exception("Unknown CommentOptionExtension")
        super().__init__(data, type_id)


########################################################
# Actual Operations
########################################################


class AccountCreate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            prefix = kwargs.pop("prefix", default_prefix)

            assert len(kwargs["new_account_name"]) <= 16, "Account name must be at most 16 chars long"

            meta = ""
            if "json_metadata" in kwargs and kwargs["json_metadata"]:
                if isinstance(kwargs["json_metadata"], dict):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]
            super().__init__(
                OrderedDict(
                    [
                        ("fee", Amount(kwargs["fee"])),
                        ("creator", String(kwargs["creator"])),
                        ("new_account_name", String(kwargs["new_account_name"])),
                        ("owner", Permission(kwargs["owner"], prefix=prefix)),
                        ("active", Permission(kwargs["active"], prefix=prefix)),
                        ("posting", Permission(kwargs["posting"], prefix=prefix)),
                        ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                        ("json_metadata", String(meta)),
                    ]
                )
            )


class AccountCreateWithDelegation(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            prefix = kwargs.pop("prefix", default_prefix)

            assert len(kwargs["new_account_name"]) <= 16, "Account name must be at most 16 chars long"

            meta = ""
            if "json_metadata" in kwargs and kwargs["json_metadata"]:
                if isinstance(kwargs["json_metadata"], dict):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]
            super().__init__(
                OrderedDict(
                    [
                        ("fee", Amount(kwargs["fee"])),
                        ("delegation", Amount(kwargs["delegation"])),
                        ("creator", String(kwargs["creator"])),
                        ("new_account_name", String(kwargs["new_account_name"])),
                        ("owner", Permission(kwargs["owner"], prefix=prefix)),
                        ("active", Permission(kwargs["active"], prefix=prefix)),
                        ("posting", Permission(kwargs["posting"], prefix=prefix)),
                        ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                        ("json_metadata", String(meta)),
                        ("extensions", Array([])),
                    ]
                )
            )


class AccountUpdate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            prefix = kwargs.pop("prefix", default_prefix)

            meta = ""
            if "json_metadata" in kwargs and kwargs["json_metadata"]:
                if isinstance(kwargs["json_metadata"], dict):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]

            owner = Permission(kwargs["owner"], prefix=prefix) if "owner" in kwargs else None
            active = Permission(kwargs["active"], prefix=prefix) if "active" in kwargs else None
            posting = Permission(kwargs["posting"], prefix=prefix) if "posting" in kwargs else None

            super().__init__(
                OrderedDict(
                    [
                        ("account", String(kwargs["account"])),
                        ("owner", Optional(owner)),
                        ("active", Optional(active)),
                        ("posting", Optional(posting)),
                        ("memo_key", PublicKey(kwargs["memo_key"], prefix=prefix)),
                        ("json_metadata", String(meta)),
                    ]
                )
            )


class AccountMetadata(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            meta = ""
            if kwargs.get("json_metadata"):
                if isinstance(kwargs["json_metadata"], dict):
                    meta = json.dumps(kwargs["json_metadata"])
                else:
                    meta = kwargs["json_metadata"]

            super().__init__(OrderedDict([("account", String(kwargs["account"])), ("json_metadata", String(meta)),]))


class Transfer(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            if "memo" not in kwargs:
                kwargs["memo"] = ""
            super().__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("amount", Amount(kwargs["amount"])),
                        ("memo", String(kwargs["memo"])),
                    ]
                )
            )


class TransferToVesting(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("amount", Amount(kwargs["amount"])),
                    ]
                )
            )


class WithdrawVesting(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [("account", String(kwargs["account"])), ("vesting_shares", Amount(kwargs["vesting_shares"])),]
                )
            )


class TransferToSavings(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            if "memo" not in kwargs:
                kwargs["memo"] = ""
            super().__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("amount", Amount(kwargs["amount"])),
                        ("memo", String(kwargs["memo"])),
                    ]
                )
            )


class TransferFromSavings(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            if "memo" not in kwargs:
                kwargs["memo"] = ""

            super().__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("request_id", Uint32(int(kwargs["request_id"]))),
                        ("to", String(kwargs["to"])),
                        ("amount", Amount(kwargs["amount"])),
                        ("memo", String(kwargs["memo"])),
                    ]
                )
            )


class CancelTransferFromSavings(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict([("from", String(kwargs["from"])), ("request_id", Uint32(int(kwargs["request_id"]))),])
            )


class ClaimRewardBalance(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("account", String(kwargs["account"])),
                        ("reward_steem", Amount(kwargs["reward_steem"])),
                        ("reward_sbd", Amount(kwargs["reward_sbd"])),
                        ("reward_vests", Amount(kwargs["reward_vests"])),
                    ]
                )
            )


class DelegateVestingShares(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("delegator", String(kwargs["delegator"])),
                        ("delegatee", String(kwargs["delegatee"])),
                        ("vesting_shares", Amount(kwargs["vesting_shares"])),
                    ]
                )
            )


class LimitOrderCreate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("owner", String(kwargs["owner"])),
                        ("orderid", Uint32(int(kwargs["orderid"]))),
                        ("amount_to_sell", Amount(kwargs["amount_to_sell"])),
                        ("min_to_receive", Amount(kwargs["min_to_receive"])),
                        ("fill_or_kill", Bool(kwargs["fill_or_kill"])),
                        ("expiration", PointInTime(kwargs["expiration"])),
                    ]
                )
            )


class LimitOrderCancel(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict([("owner", String(kwargs["owner"])), ("orderid", Uint32(int(kwargs["orderid"]))),])
            )


class SetWithdrawVestingRoute(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("from_account", String(kwargs["from_account"])),
                        ("to_account", String(kwargs["to_account"])),
                        ("percent", Uint16((kwargs["percent"]))),
                        ("auto_vest", Bool(kwargs["auto_vest"])),
                    ]
                )
            )


class Convert(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("owner", String(kwargs["owner"])),
                        ("requestid", Uint32(kwargs["requestid"])),
                        ("amount", Amount(kwargs["amount"])),
                    ]
                )
            )


class FeedPublish(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("publisher", String(kwargs["publisher"])),
                        ("exchange_rate", ExchangeRate(kwargs["exchange_rate"])),
                    ]
                )
            )


class WitnessUpdate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            prefix = kwargs.pop("prefix", default_prefix)

            if not kwargs["block_signing_key"]:
                kwargs["block_signing_key"] = "GLS1111111111111111111111111111111114T1Anm"
            super().__init__(
                OrderedDict(
                    [
                        ("owner", String(kwargs["owner"])),
                        ("url", String(kwargs["url"])),
                        ("block_signing_key", PublicKey(kwargs["block_signing_key"], prefix=prefix)),
                        ("props", ChainProperties(kwargs["props"])),
                        ("fee", Amount(kwargs["fee"])),
                    ]
                )
            )


class ChainPropertiesUpdate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            props = kwargs.get("props")

            # A hack to extract properties at the second op processing in transactionbuilder
            if props and type(props) == list:
                props = props[1]

            if props and type(props) == dict:
                type_id = 0
                if "min_delegation" in props:
                    type_id = 1
                if "auction_window_size" in props:
                    type_id = 2
                if "worker_reward_percent" in props:
                    type_id = 3
                if "claim_idleness_time" in props:
                    type_id = 4

                obj = [type_id, {"props": props}]
                props = Props(obj)

            super().__init__(OrderedDict([("owner", String(kwargs["owner"])), ("props", props),]))


class AccountWitnessVote(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("account", String(kwargs["account"])),
                        ("witness", String(kwargs["witness"])),
                        ("approve", Bool(bool(kwargs["approve"]))),
                    ]
                )
            )


class CustomJson(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            if "json" in kwargs and kwargs["json"]:
                if isinstance(kwargs["json"], dict) or isinstance(kwargs["json"], list):
                    js = json.dumps(kwargs["json"])
                else:
                    js = kwargs["json"]

            if len(kwargs["id"]) > 32:
                raise Exception("'id' too long")

            super().__init__(
                OrderedDict(
                    [
                        ("required_auths", Array([String(o) for o in kwargs["required_auths"]])),
                        ("required_posting_auths", Array([String(o) for o in kwargs["required_posting_auths"]])),
                        ("id", String(kwargs["id"])),
                        ("json", String(js)),
                    ]
                )
            )


class CommentOptions(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            new_extensions = []

            # handle beneficiaries
            if "beneficiaries" in kwargs and kwargs["beneficiaries"]:
                new_extensions.append([0, {"beneficiaries": kwargs["beneficiaries"]}])

            if "auction_window_reward_destination" in kwargs and kwargs["auction_window_reward_destination"]:
                new_extensions.append([1, {"destination": kwargs["auction_window_reward_destination"]}])

            if "curation_rewards_percent" in kwargs and kwargs["curation_rewards_percent"]:
                new_extensions.append([2, {"percent": kwargs["curation_rewards_percent"]}])

            if new_extensions:
                kwargs["extensions"] = new_extensions

            extensions = Array([])
            if "extensions" in kwargs and kwargs["extensions"]:
                extensions = Array([CommentOptionExtensions(o) for o in kwargs["extensions"]])

            super().__init__(
                OrderedDict(
                    [
                        ("author", String(kwargs["author"])),
                        ("permlink", String(kwargs["permlink"])),
                        ("max_accepted_payout", Amount(kwargs["max_accepted_payout"])),
                        ("percent_steem_dollars", Uint16(int(kwargs["percent_steem_dollars"]))),
                        ("allow_votes", Bool(bool(kwargs["allow_votes"]))),
                        ("allow_curation_rewards", Bool(bool(kwargs["allow_curation_rewards"]))),
                        ("extensions", extensions),
                    ]
                )
            )


class ProposalCreate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            assert kwargs["proposed_operations"], "proposed_operations cannot be empty!"

            if isinstance(kwargs["proposed_operations"][0], GrapheneObject):
                proposed_operations = [OperationWrapper(Operation(op)) for op in kwargs["proposed_operations"]]
            else:
                proposed_operations = [OperationWrapper(Operation(op["op"])) for op in kwargs["proposed_operations"]]

            review_period_time = PointInTime(kwargs["review_period_time"]) if kwargs.get("review_period_time") else None

            super().__init__(
                OrderedDict(
                    [
                        ("author", String(kwargs["author"])),
                        ("title", String(kwargs["title"])),
                        ("memo", String(kwargs.get("memo", ""))),
                        ("expiration_time", PointInTime(kwargs["expiration_time"])),
                        ("proposed_operations", Array(proposed_operations)),
                        ("review_period_time", Optional(review_period_time)),
                        ("extensions", Array(kwargs.get("extensions") or [])),
                    ]
                )
            )


class ProposalUpdate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            active_approvals_to_add = [String(str(x)) for x in kwargs.get("active_approvals_to_add") or []]
            active_approvals_to_remove = [String(str(x)) for x in kwargs.get("active_approvals_to_remove") or []]
            owner_approvals_to_add = [String(str(x)) for x in kwargs.get("owner_approvals_to_add") or []]
            owner_approvals_to_remove = [String(str(x)) for x in kwargs.get("owner_approvals_to_remove") or []]
            posting_approvals_to_add = [String(str(x)) for x in kwargs.get("posting_approvals_to_add") or []]
            posting_approvals_to_remove = [String(str(x)) for x in kwargs.get("posting_approvals_to_remove") or []]
            key_approvals_to_add = [String(str(x)) for x in kwargs.get("key_approvals_to_add") or []]
            key_approvals_to_remove = [String(str(x)) for x in kwargs.get("key_approvals_to_remove") or []]

            super().__init__(
                OrderedDict(
                    [
                        ("author", String(kwargs["author"])),
                        ("title", String(kwargs["title"])),
                        ("active_approvals_to_add", Array(active_approvals_to_add)),
                        ("active_approvals_to_remove", Array(active_approvals_to_remove)),
                        ("owner_approvals_to_add", Array(owner_approvals_to_add)),
                        ("owner_approvals_to_remove", Array(owner_approvals_to_remove)),
                        ("posting_approvals_to_add", Array(posting_approvals_to_add)),
                        ("posting_approvals_to_remove", Array(posting_approvals_to_remove)),
                        ("key_approvals_to_add", Array(key_approvals_to_add)),
                        ("key_approvals_to_remove", Array(key_approvals_to_remove)),
                        ("extensions", Array(kwargs.get("extensions") or [])),
                    ]
                )
            )


class ProposalDelete(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("author", String(kwargs["author"])),
                        ("title", String(kwargs["title"])),
                        ("requester", String(kwargs["requester"])),
                        ("extensions", Array(kwargs.get("extensions") or [])),
                    ]
                )
            )


class Claim(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            super().__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("amount", Amount(kwargs["amount"])),
                        ("to_vesting", Bool(bool(kwargs["to_vesting"]))),
                        ("extensions", Array([])),
                    ]
                )
            )


class VariantObject(GrapheneObject):
    """
    Represents fc::variant_object.

    variant_object is a dict-like structure with keys and values, where keys are always strings, and values could be
    anything. We provide here only limited set of values recognition.
    """

    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            self.data = OrderedDict()
            for key, value in kwargs.items():
                if isinstance(value, str):
                    _value = String(value)
                elif isinstance(value, int):
                    if value > 0:
                        _value = Uint64(value)
                    else:
                        _value = Int64(value)

                self.data[key] = _value

    def __bytes__(self):
        if self.data is None:
            return bytes()

        # Encode number of elements
        b = varint(len(self.data))
        for name, value in self.data.items():
            # All keys are strings
            b += bytes(String(name))

            if isinstance(value, String):
                b += b"\x05"  # delimiter
            elif isinstance(value, Int64):
                b += b"\x01"
            else:
                b += b"\x02"
            b += bytes(value)
        return b


class DonateMemo(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            comment = String(kwargs["comment"]) if kwargs.get("comment") else None

            super().__init__(
                OrderedDict(
                    [
                        ("app", String(kwargs["app"])),
                        ("version", Uint16(int(kwargs["version"]))),
                        ("target", kwargs["target"]),
                        ("comment", Optional(comment)),
                    ]
                )
            )


class Donate(GrapheneObject):
    def __init__(self, *args, **kwargs):
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]

            super().__init__(
                OrderedDict(
                    [
                        ("from", String(kwargs["from"])),
                        ("to", String(kwargs["to"])),
                        ("amount", Amount(kwargs["amount"])),
                        ("memo", kwargs["memo"]),
                        ("extensions", Array([])),
                    ]
                )
            )


def isArgsThisClass(self, args):
    return len(args) == 1 and type(args[0]).__name__ == type(self).__name__



import json
import logging
import re
from datetime import datetime
from queue import Queue

from funcy import get_in, silent, walk_values

from golos.amount import Amount
from golos.commit import Commit
from golos.instance import shared_steemd_instance
from golos.utils import (
    calculate_hot,
    calculate_trending,
    construct_identifier,
    parse_time,
    remove_from_dict,
    resolve_identifier,
)
from golosbase.exceptions import PostDoesNotExist, RPCError, VotingInvalidOnArchivedPost
from golosbase.operations import CommentOptions

log = logging.getLogger(__name__)


class Post(dict):
    """
    This object gets instantiated by Steem.streams and is used as an abstraction layer for Comments in Steem.

    Args:
        post (str or dict): ``@author/permlink`` or raw ``comment`` as dictionary.
        steemd_instance (Steemd): Steemd node to connect to
        vote_limit (int): limit number of upvotes
    """

    def __init__(self, post, steemd_instance=None, vote_limit=10000):
        self.steemd = steemd_instance or shared_steemd_instance()
        self.vote_limit = vote_limit
        self.commit = Commit(steemd_instance=self.steemd)

        # will set these during refresh()
        self.patched = False
        self.category = None
        self.root_identifier = None

        if isinstance(post, str):  # From identifier
            self.identifier = self.parse_identifier(post)
        elif isinstance(post, dict) and "author" in post and "permlink" in post:
            self.identifier = construct_identifier(post["author"], post["permlink"])
        else:
            raise ValueError("Post expects an identifier or a dict " "with author and permlink!")

        self.refresh()

    @staticmethod
    def parse_identifier(uri):
        """Extract canonical post id/url (i.e. strip any leading `@`)."""
        return "@%s" % uri.split("@")[-1]

    def refresh(self):
        post_author, post_permlink = resolve_identifier(self.identifier)
        post = self.steemd.get_content(post_author, post_permlink, vote_limit=self.vote_limit)
        if not post["permlink"]:
            raise PostDoesNotExist("Post does not exist: %s" % self.identifier)

        # If this 'post' comes from an operation, it might carry a patch
        if "body" in post and re.match("^@@", post["body"]):
            self.patched = True

        # TODO: Check
        # This field is returned from blockchain, but it's empty. Fill it
        try:
            reblogged_by = [i for i in self.steemd.get_reblogged_by(post_author, post_permlink) if i != post_author]
        except RPCError:
            reblogged_by = []

        post["reblogged_by"] = reblogged_by

        # Parse Times
        parse_times = ["active", "cashout_time", "created", "last_payout", "last_update", "max_cashout_time"]
        for p in parse_times:
            post[p] = parse_time(post.get(p, "1970-01-01T00:00:00"))

        # Parse Amounts
        sbd_amounts = [
            "total_payout_value",
            "max_accepted_payout",
            "pending_payout_value",
            "curator_payout_value",
            "total_pending_payout_value",
            "promoted",
        ]
        for p in sbd_amounts:
            post[p] = Amount(post.get(p, "0.000 GBG"))

        # calculate trending and hot scores for sorting
        post["score_trending"] = calculate_trending(post.get("net_rshares", 0), post["created"])
        post["score_hot"] = calculate_hot(post.get("net_rshares", 0), post["created"])

        # turn json_metadata into python dict
        meta_str = post.get("json_metadata", "{}")
        post["json_metadata"] = silent(json.loads)(meta_str) or {}

        post["tags"] = []
        post["community"] = ""
        if isinstance(post["json_metadata"], dict):
            if post["depth"] == 0:
                tags = [post["parent_permlink"]]
                tags += get_in(post, ["json_metadata", "tags"], default=[])
                tags_added = set()
                post["tags"] = [tag for tag in tags if not (tag in tags_added or tags_added.add(tag))]

            post["community"] = get_in(post, ["json_metadata", "community"], default="")

        # If this post is a comment, retrieve the root comment
        self.root_identifier, self.category = self._get_root_identifier(post)

        self._store_post(post)

    def _store_post(self, post):
        # Store original values as obtained from the rpc
        for key, value in post.items():
            super(Post, self).__setitem__(key, value)

        # Set attributes as well
        for key in post:
            setattr(self, key, post[key])

        # also set identifier
        super(Post, self).__setitem__("identifier", self.identifier)

    def __getattr__(self, key):
        return object.__getattribute__(self, key)

    def __getitem__(self, key):
        return super(Post, self).__getitem__(key)

    def __repr__(self):
        return "<Post-%s>" % self.identifier

    __str__ = __repr__

    def _get_root_identifier(self, post=None):
        if not post:
            post = self
        m = re.match("/([^/]*)/@([^/]*)/([^#]*).*", post.get("url", ""))
        if not m:
            return "", ""
        else:
            category = m.group(1)
            author = m.group(2)
            permlink = m.group(3)
            return construct_identifier(author, permlink), category

    def get_replies(self):
        """Return **first-level** comments of the post."""
        post_author, post_permlink = resolve_identifier(self.identifier)
        replies = self.steemd.get_content_replies(post_author, post_permlink)
        return map(silent(Post), replies)

    @staticmethod
    def get_all_replies(root_post=None):
        """
        Recursively fetch all the child comments, and return them as a list.

        Usage: all_comments = Post.get_all_replies(Post('@foo/bar'))
        """
        queue = Queue()
        replies = []
        [queue.put(reply) for reply in root_post.get_replies()]
        while not queue.empty():
            reply = queue.get()
            if reply:
                [queue.put(children) for children in reply.get_replies()]
                replies.append(reply)
        return replies

    @property
    def reward(self):
        """Return a float value of estimated total SBD reward."""
        return Amount(self.get("total_payout_value", "0 SBD")) + Amount(self.get("pending_payout_value", "0 SBD"))

    def time_elapsed(self):
        """Return a timedelta on how old the post is."""
        return datetime.utcnow() - self["created"]

    def is_main_post(self):
        """Retuns True if main post, and False if this is a comment (reply)."""
        return self["depth"] == 0

    def is_comment(self):
        """Retuns True if post is a comment."""
        return self["depth"] > 0

    def curation_reward_pct(self):
        """If post is less than 30 minutes old, it will incur a curation reward penalty."""
        reward = (self.time_elapsed().seconds / 1800) * 100
        if reward > 100:
            reward = 100
        return reward

    def export(self):
        """This method returns a dictionary that is type-safe to store as JSON or in a database."""
        self.refresh()

        # Remove Steem instance object
        safe_dict = remove_from_dict(self, ["steemd", "commit"])

        # Convert Amount class objects into pure dictionaries
        def decompose_amounts(item):
            if type(item) == Amount:
                return dict(item)
            return item

        return walk_values(decompose_amounts, safe_dict)

    ######################
    # Commital Properties
    ######################
    def upvote(self, weight=+100, voter=None):
        """
        Upvote the post.

        :param float weight: (optional) Weight for posting (-100.0 - +100.0) defaults to +100.0
        :param str voter: (optional) Voting account
        """
        return self.vote(weight, voter=voter)

    def downvote(self, weight=-100, voter=None):
        """
        Downvote the post.

        :param float weight: (optional) Weight for posting (-100.0 - +100.0) defaults to -100.0
        :param str voter: (optional) Voting account
        """
        return self.vote(weight, voter=voter)

    def vote(self, weight, voter=None):
        """
        Vote the post.

        :param float weight: Weight for posting (-100.0 - +100.0)
        :param str voter: Voting account
        """
        # Test if post is archived, if so, voting is worthless but just
        # pollutes the blockchain and account history
        if self.is_main_post() and self.get("net_rshares") is None:
            raise VotingInvalidOnArchivedPost
        return self.commit.vote(self.identifier, weight, account=voter)

    def edit(self, body, meta=None, replace=False):
        """
        Edit an existing post.

        :param str body: Body of the reply
        :param json meta: JSON meta object that can be attached to the
                          post. (optional)
        :param bool replace: Instead of calculating a *diff*, replace
                             the post entirely (defaults to ``False``)
        """
        if not meta:
            meta = {}
        original_post = self

        if replace:
            newbody = body
        else:
            import diff_match_patch

            dmp = diff_match_patch.diff_match_patch()
            patch = dmp.patch_make(original_post["body"], body)
            newbody = dmp.patch_toText(patch)

            if not newbody:
                log.info("No changes made! Skipping ...")
                return

        reply_identifier = construct_identifier(original_post["parent_author"], original_post["parent_permlink"])

        new_meta = {}
        if meta:
            if original_post["json_metadata"]:
                pass

                new_meta = original_post["json_metadata"].update(meta)
            else:
                new_meta = meta

        return self.commit.post(
            original_post["title"],
            newbody,
            reply_identifier=reply_identifier,
            author=original_post["author"],
            permlink=original_post["permlink"],
            json_metadata=new_meta,
        )

    def reply(self, body, title="", author="", meta=None):
        """
        Reply to an existing post.

        :param str body: Body of the reply
        :param str title: Title of the reply post
        :param str author: Author of reply (optional) if not provided
                           ``default_user`` will be used, if present, else
                           a ``ValueError`` will be raised.
        :param json meta: JSON meta object that can be attached to the
                          post. (optional)
        """
        return self.commit.post(title, body, json_metadata=meta, author=author, reply_identifier=self.identifier)

    def set_comment_options(self, options):
        op = CommentOptions(
            **{
                "author": self["author"],
                "permlink": self["permlink"],
                "max_accepted_payout": options.get("max_accepted_payout", str(self["max_accepted_payout"])),
                "percent_steem_dollars": int(
                    options.get("percent_steem_dollars", self["percent_steem_dollars"] / 100) * 100
                ),
                "allow_votes": options.get("allow_votes", self["allow_votes"]),
                "allow_curation_rewards": options.get("allow_curation_rewards", self["allow_curation_rewards"]),
                "extensions": options.get("extensions", []),
            }
        )
        return self.commit.finalizeOp(op, self["author"], "posting")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['golos', 'golosbase']

package_data = \
{'': ['*']}

install_requires = \
['certifi>=2020.4.5,<2021.0.0',
 'ecdsa>=0.13,<0.14',
 'funcy>=1.14,<2.0',
 'langdetect>=1.0.8,<2.0.0',
 'prettytable>=0.7.2,<0.8.0',
 'pycryptodome>=3.9.7,<4.0.0',
 'pylibscrypt>=1.8.0,<2.0.0',
 'scrypt>=0.8.13,<0.9.0',
 'toolz>=0.10.0,<0.11.0',
 'ujson>=2.0.3,<3.0.0',
 'urllib3>=1.25.9,<2.0.0',
 'voluptuous>=0.11.7,<0.12.0',
 'w3lib>=1.21.0,<2.0.0',
 'websocket-client>=0.56,<0.57']

setup_kwargs = {
    'name': 'python-golos',
    'version': '1.1.1',
    'description': 'Python library for Golos blockchain',
    'long_description': '# Python GOLOS Library\n\nThis is a fork of [golodranets](https://github.com/steepshot/golodranets) GOLOS\nlibrary for Python which was forked from official STEEM library for Python. It\ncomes with a BIP38 encrypted wallet.\n\nThe main differences from the `steem-python`:\n\n* directed to work with GOLOS blockchain\n* websocket support\n* convert Cyrillic to Latin for tags and categories\n* Golos assets - `STEEM` -> `GOLOS`, `SBD` -> `GBG`, `VESTS` -> `GESTS`\n* renamed modules - `steem` -> `golos`, `steemdata` -> `golosdata`\n* for `Post` instance added two fields - `score_trending` and `score_hot`. This fields may be helpful if you want to sort your saved posts like `get_discussions_by_trending` and `get_discussions_by_trending` methods do. `reblogged_by` field is also filled now\n* for `Account` instance methods `get_followers` and `get_following` were improved - now it takes `limit` and `offset` parameters\n\nGOLOS HF 23 is supported.\n\n# Installation\n\nAs regular package:\n\n```\npip install python-golos\n```\n\nLocal installation from source:\n\n```\ncd golos-python\npoetry install\npoetry shell\n```\n\n## Homebrew Build Prereqs\n\nIf you\'re on a mac, you may need to do the following first:\n\n```\nbrew install openssl\nexport CFLAGS="-I$(brew --prefix openssl)/include $CFLAGS"\nexport LDFLAGS="-L$(brew --prefix openssl)/lib $LDFLAGS"\n```\n',
    'author': 'Vladimir Kamarzin',
    'author_email': 'vvk@vvk.pp.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bitfag/golos-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
